## main group
EDGE = "edge_features"

## metafeatures
NAME = "_name"
INDEX = "_index"

## generic features
DISTANCE = "distance"  # float; former FEATURENAME_EDGEDISTANCE
SAMECHAIN = "same_chain"  # bool; former FEATURE_EDGE_INTERACTIONTYPE & FEATURENAME_EDGESAMECHAIN
SAMERES = "same_res"  # bool

## interactions
COVALENT = "covalent"  # bool; former FEATURENAME_COVALENT
ELEC = "electrostatic"  # float; former FEATURENAME_EDGECOULOMB
VDW = "vanderwaals"  # float; former FEATURENAME_EDGEVANDERWAALS
