# list of default targets and tasks in hdf5 files

## main group
VALUES = "target_values"

## classification tasks
BINARY = "binary"
CAPRI = "capri_class"

## regression tasks
IRMSD = "irmsd"
LRMSD = "lrmsd"
FNAT = "fnat"
DOCKQ = "dockq"

## task names
REGRESS = "regress"
CLASSIF = "classif"
